#coding: utf-8
'''

This software is licensed under the following agreement:

JHU Academic Software License Agreement

This license agreement ("License"), effective automatically by and between you (hereinafter referred to as the "Licensee") and The Johns Hopkins University (“JHU”) for use of the software known as CHASM (the “Software”) once and only so long as Licensee complies with the following terms and conditions:

   1. The requirement to acknowledge the copyright of JHU as follows: “Copyright Johns Hopkins University 2010, all rights reserved,” and copyrights of any incorporated third party software as described in the associated documentation.
   2. Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
         1. Redistributions of source code must retain the above copyright notice, and these terms and conditions.
         2. Neither the name of JHU nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission from an authorized JHU representative.
         3. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. LICENSEE AGREES TO DEFEND, INDEMNIFY AND HOLD HARMLESS JHU FOR ANY CLAIMS ARISING FROM LICENSEE’S USE OF THE SOFTWARE TO THE FULLEST EXTENT PERMITTED BY LAW.
   3. Only a non-exclusive, nontransferable license is granted to Licensee to use the Software for academic, non-profit, or government-sponsored research purposes. Use of the Software under this License is restricted to non-commercial purposes. Commercial use of the Software requires a separately executed written license agreement.
   4. Licensee agrees that it will use the Software, and any modifications, improvements, or derivatives to the Software that the Licensee may create (collectively, "Improvements") solely for non-commercial purposes and shall not distribute or transfer the Software or Improvements to any person or third parties without requiring that such third parties adhere to the terms of this License.
   5. Licensee acknowledges that JHU holds copyright in the Software or portions of the Software, and that the Software incorporates third party software which may be subject to additional terms and conditions. Licensee agrees that any Improvements made by Licensee shall be subject to the same terms and conditions as the Software.
   6. Licensee agrees that any publication of results obtained with the Software will acknowledge its use by an appropriate citation.
   7. Licensee’s rights under this License terminate automatically without notice from JHU if you fail to comply with any term(s) of this License.
   8. This License shall be governed by the laws of the State of Maryland, excluding the application of its conflicts of law rules. Licensee agrees that any dispute shall be appropriate only in the state and federal courts located within the State of Maryland.

Created Nov 10, 2011

Contact: chasm-users@lists.johnshopkins.edu
'''

import math
import sys
import random

# external imports
from DataSet import Mutation

# vars
rcDict = {"A":"T","T":"A","C":"G","G":"C","N":"N","a":"t","t":"a","g":"c","c":"g","n":"n"}
codonTable = {"ATG":"M", "GCT":"A", "GCC":"A", "GCA":"A", "GCG":"A", "TGT":"C", "TGC":"C", "GAT":"D", "GAC":"D", "GAA":"E", "GAG":"E", "TTT":"F", "TTC":"F", "GGT":"G", "GGC":"G", "GGA":"G", "GGG":"G", "CAT":"H", "CAC":"H", "ATT":"I", "ATC":"I", "ATA":"I", "AAA":"K", "AAG":"K", "TTA":"L", "TTG":"L", "CTT":"L", "CTC":"L", "CTA":"L", "CTG":"L", "AAT":"N", "AAC":"N", "CCT":"P", "CCC":"P", "CCA":"P", "CCG":"P", "CAA":"Q", "CAG":"Q", "TCT":"S", "TCC":"S", "TCA":"S", "TCG":"S", "AGT":"S", "AGC":"S", "ACT":"T", "ACC":"T", "ACA":"T", "ACG":"T", "CGT":"R", "CGC":"R", "CGA":"R", "CGG":"R", "AGA":"R", "AGG":"R", "GTT":"V", "GTC":"V", "GTA":"V", "GTG":"V", "TGG":"W", "TAT":"Y", "TAC":"Y", "TGA":"*", "TAA":"*", "TAG":"*","AUG":"M", "GCU":"A", "GCC":"A", "GCA":"A", "GCG":"A", "UGU":"C", "UGC":"C", "GAU":"D", "GAC":"D", "GAA":"E", "GAG":"E", "UUU":"F", "UUC":"F", "GGU":"G", "GGC":"G", "GGA":"G", "GGG":"G", "CAU":"H", "CAC":"H", "AUU":"I", "AUC":"I", "AUA":"I", "AAA":"K", "AAG":"K", "UUA":"L", "UUG":"L", "CUU":"L", "CUC":"L", "CUA":"L", "CUG":"L", "AAU":"N", "AAC":"N", "CCU":"P", "CCC":"P", "CCA":"P", "CCG":"P", "CAA":"Q", "CAG":"Q", "UCU":"S", "UCC":"S", "UCA":"S", "UCG":"S", "AGU":"S", "AGC":"S", "ACU":"T", "ACC":"T", "ACA":"T", "ACG":"T", "CGU":"R", "CGC":"R", "CGA":"R", "CGG":"R", "AGA":"R", "AGG":"R", "GUU":"V", "GUC":"V", "GUA":"V", "GUG":"V", "UGG":"W", "UAU":"Y", "UAC":"Y", "UGA":"*", "UAA":"*", "UAG":"*"}
base_rev_dic = {'A':'T', 'T':'A', 'G':'C', 'C':'G', 'X':'X', '.':'.', 'N':'N', 'a':'t', 't':'a', 'g':'c', 'c':'g', 'x':'x', 'n':'n', '-':'-', '':''}

class CodonConverter(object):
    """
    Convert genomic coordinates and a nucleotide substitution into
    a transcript and amino acid substitution.

    Note: Only missense variation will be returned unless missenseOnly is set to False.
    """

    def __init__(self, dbconn, \
                 missense=False, \
                 inframe_indel=False, \
                 frameshift_indel=False, \
                 pickOneTranscript=False):
        self.conn = dbconn
        self.missense = missense
        self.inframe_indel = inframe_indel
        self.frameshift_indel = frameshift_indel
        self.pickOneTranscript = pickOneTranscript
        #self.codon_display_len_cutoff = 40
        self.line_count = None


    def convert(self, genomic, line_count=None):
        """Get codon, make mutation, return transcript + aa subst if mutation is missense """

        try:
            self.line_count = line_count

            vars = []
            best_var = None

            (codons, best_codon, no_transcripts_for_genomic, ref_match_for_uid) = self.__get_codons(genomic)

            if ref_match_for_uid == True:

                if best_codon != None:
                    best_var = Mutation('\t'.join([genomic.uid, \
                                                   best_codon.mrnaAcc, \
                                                   best_codon.codon_change, \
                                                   str(best_codon.codonNo), \
                                                   best_codon.ref_aa, \
                                                   best_codon.alt_aa, \
                                                   str(best_codon.aa_len), \
                                                   best_codon.strand \
                                                  ]))
                    best_var.UID = best_codon.UID;
                    best_var.storeGenomicCoords(genomic)
                    best_var.sequence_ontology = best_codon.sequence_ontology
                    best_var.transcript_count = no_transcripts_for_genomic

                for codon in codons:

                    var = None

                    if codon != None:

                        # Checks if the reference base matches.
                        if codon.sequence_ontology in ['MS', 'SY', 'SG', 'SL'] and \
                                self.__matches(genomic, codon) == False:
                            continue

                        addFlag = False
                        if self.missense and codon.sequence_ontology == 'MS':
                            addFlag = True
                        elif self.frameshift_indel and codon.sequence_ontology in ['FI', 'FD']:
                            addFlag = True
                        elif self.inframe_indel and codon.sequence_ontology in ['II', 'ID']:
                            addFlag = True
                        elif self.missense == False and self.frameshift_indel == False \
                                and self.inframe_indel == False:
                            addFlag = True
                        if addFlag:
                            var = Mutation("\t".join([genomic.uid, \
                                                codon.mrnaAcc, \
                                                codon.codon_change, \
                                                str(codon.codonNo), \
                                                codon.ref_aa, \
                                                codon.alt_aa, \
                                                str(codon.aa_len), \
                                                codon.strand]))
                            var.UID = codon.UID;
                            var.storeGenomicCoords(genomic)
                            var.sequence_ontology = codon.sequence_ontology
                            var.transcript_count = no_transcripts_for_genomic
                    else:
                        pass

                    if var != None:
                        vars.append(var)

            return (vars, best_var, ref_match_for_uid)
        except Exception, e:
            raise

    def __get_reverse_seq (self, seq):
        try:
            new_seq = ''
            for i in xrange(len(seq)-1, -1, -1):
                new_seq += base_rev_dic[seq[i]]
            return new_seq
        except Exception, e:
            raise

    def __get_ref_codon_bases (self, codon, genomic):
        try:
            if codon.strand == genomic.strand:
                start_portion = codon.nucs[:codon.codon_pos].lower()
                if genomic.ref == '-':
                    middle_portion = ''
                else:
                    middle_portion = genomic.ref.upper()
                codon_display = start_portion + middle_portion
                len_codon_display = len(codon_display)
                len_codon_display_3_div_rem = len_codon_display % 3
                if len_codon_display <= 3:
                    last_portion = codon.nucs[len_codon_display:].lower()
                else:
                    if len_codon_display_3_div_rem == 0:
                        last_portion = ''
                    else:
                        codon_no_increase = len_codon_display / 3
                        last_codon_bases = CodonTblRow.get_codon_bases(self.conn, codon.UID, codon.codonNo + codon_no_increase)
                        if last_codon_bases == None:
                            last_portion = '???'
                        else:
                            last_portion = last_codon_bases[len_codon_display_3_div_rem:].lower()
                codon_display += last_portion
                return codon_display, codon.codonNo, start_portion, middle_portion, last_portion
            else:
                if genomic.ref == '-':
                    middle_portion = ''
                else:
                    rev_change_bases = self.__get_reverse_seq(genomic.ref)
                    middle_portion = rev_change_bases.upper()
                if genomic.ref == '-':
                    last_portion = codon.nucs[codon.codon_pos:].lower()
                else:
                    last_portion = codon.nucs[codon.codon_pos + 1:].lower()
                codon_display = middle_portion + last_portion
                len_codon_display = len(codon_display)
                len_codon_display_3_div_rem = len_codon_display % 3
                if len_codon_display <= 3:
                    first_codon_no = codon.codonNo
                    if len_codon_display_3_div_rem == 0:
                        first_portion = ''
                    else:
                        first_portion = codon.nucs[:3 - len_codon_display_3_div_rem].lower()
                else:
                    if len_codon_display_3_div_rem == 0:
                        codon_no_decrease = len_codon_display / 3 - 1
                        first_codon_no = codon.codonNo - codon_no_decrease
                        first_portion = ''
                    else:
                        codon_no_decrease = len_codon_display / 3
                        first_codon_no = codon.codonNo - codon_no_decrease
                        first_codon_bases = CodonTblRow.get_codon_bases(self.conn, codon.UID, first_codon_no)
                        if first_codon_bases == None:
                            first_portion = '???'
                        else:
                            first_portion = first_codon_bases[:3 - len_codon_display_3_div_rem].lower()
                codon_display = first_portion + codon_display
                return codon_display, first_codon_no, first_portion, middle_portion, last_portion
        except Exception, e:
            raise

    #     This method will check if the reference input is the correct base. The first position in the reference base input
    #     will be compared to the base at that position in the database.
    def __check_if_ref_input_correct(self, genomic, codon):

        try:
            input_ref_error = 0
            input_ref = genomic.ref
            actual_ref = codon.nucs[codon.codon_pos]
            reversed_actual_ref = self.__get_reverse_seq(actual_ref)
            if input_ref != "-":
                ref_first_pos = input_ref[0]
                if genomic.strand == codon.strand:
                    if ref_first_pos == actual_ref:
                        pass
                    else:
                        input_ref_error = 1
                else:
                    if ref_first_pos == reversed_actual_ref:
                        pass
                    else:
                        input_ref_error = 1

            input_ref_correct = True        #By default this is true but will be changed to False if the reference base does not match the base in the database at the chromosome and position
            if input_ref_error != 0:
                input_ref_correct = False
                cursor = self.conn.cursor()
                stmt = 'select GeneSymbol from GeneSymbols where UID=' + str(codon.UID)
                cursor.execute(stmt)
                result = cursor.fetchone()
                if result == None:
                    hugo = 'Unknown'
                else:
                    hugo = result[0]
                if input_ref_error == 1:
                    error_msg = 'Input reference base is incorrect for the specified chromosome and position.'
                    if genomic.strand != codon.strand:
                        error_msg += ' Gene is on the opposite strand to the input strand.'
                if hasattr(genomic, 'input_line_no'):
                    input_line_no = genomic.input_line_no
                else:
                    input_line_no = 'NA'
                sys.stderr.write('$%$'.join([
                                             input_line_no,
                                             genomic.uid,
                                             hugo,
                                             error_msg,
                                             str(genomic)
                                             ]
                                            ) + '\n')
            return input_ref_correct
        except Exception, tg:
            raise

    def __set_ref_alt_codons(self, codon, genomic):
        try:
            codon.ref_codon, \
                codon.codonNo, \
                ref_codon_first_portion, \
                ref_codon_middle_portion, \
                ref_codon_last_portion = self.__get_ref_codon_bases(codon, genomic)

            if genomic.alt == '-':
                alt_seq = ''
            else:
                if codon.strand != genomic.strand:
                    alt_seq = self.__get_reverse_seq(genomic.alt)
                else:
                    alt_seq = genomic.alt
            codon.alt_codon = ref_codon_first_portion + alt_seq.upper() + ref_codon_last_portion

#             Kyle Moad 4/3/2017: ref/alt_codon display cutoff interferes with protein consequence prediction
#             if len(codon.ref_codon) > self.codon_display_len_cutoff:
#                 codon.ref_codon = str(len(codon.ref_codon)) + 'nt'
#             if len(codon.alt_codon) > self.codon_display_len_cutoff:
#                 codon.alt_codon = str(len(codon.alt_codon)) + 'nt'
            codon.codon_change = codon.ref_codon + '>' + codon.alt_codon
        except Exception, e:
            raise

    def __set_ref_alt_aas(self, codon, genomic):
        try:
            pass
            ref_len = len(codon.ref_codon)
            alt_len = len(codon.alt_codon)
            codon.ref_aa = self.__translate_codons(codon.ref_codon) # Will ref_codon ever not be a multiple of 3?
            if alt_len%3 == 0:
                codon.alt_aa = self.__translate_codons(codon.alt_codon)
            else:
                codon.alt_aa = self.__get_aas_until_stop(codon)
            return codon
        except Exception, e:
            raise

    def __get_aas_until_stop(self, codon):
        try:
            aas, extra_bases = self.__translate_bases(codon.alt_codon)
            stop_found = '*' in aas
            end_of_transcript = False
            if not(stop_found):
                chunk_len = 20
                for chunk_bases, complete_chunk in CodonTblRow.get_bases_in_chunks(self.conn, codon.UID, codon.codonNo + 1, chunk_len):
                    full_bases = extra_bases + chunk_bases
                    chunk_aas, extra_bases = self.__translate_bases(full_bases)
                    aas += chunk_aas
                    stop_found = '*' in chunk_aas
                    if stop_found:
                        break
                end_of_transcript = not(complete_chunk)
            if stop_found:
                aas = aas[:aas.index('*')+1]
            elif end_of_transcript:
                aas += '$'
            return aas
        except Exception, e:
            raise

    def __get_deletion_sequence_ontology(self, transcript_uid, chrom, genomic_strand, deletion_start_pos, deletion_size):
        if genomic_strand == '+':
            deletion_end_pos = deletion_start_pos + deletion_size - 1
        else:
            deletion_end_pos = deletion_start_pos
            deletion_start_pos = deletion_end_pos - deletion_size + 1

        cursor = self.conn.cursor()

        stmt = 'select min(pos1), min(pos3), max(pos1), max(pos3) from CodonTable where UID=' + transcript_uid
        cursor.execute(stmt)
        (min_pos1, min_pos3, max_pos1, max_pos3) = cursor.fetchone()
        coding_min_pos = min(min_pos1, min_pos3)
        coding_max_pos = max(max_pos1, max_pos3)

        deletion_start_pos_str = str(deletion_start_pos)
        deletion_end_pos_str = str(deletion_end_pos)
        stmt = 'select cStart, cEnd from Transcript_Exon ' +\
               'where UID=' + transcript_uid + ' ' +\
               'and Chr="' + chrom + '" ' +\
               'and ((cStart<=' + deletion_start_pos_str + ' and cEnd>=' + deletion_start_pos_str + ') ' +\
               'or (cStart<=' + deletion_end_pos_str + ' and cEnd>=' + deletion_end_pos_str + '))'
        cursor.execute(stmt)
        results = cursor.fetchall()

        ds = deletion_start_pos
        de = deletion_end_pos
        ts = coding_min_pos
        te = coding_max_pos

        number_of_nt_deleted_in_exons = 0
        for result in results:
            (cs, ce) = result
            cde = min(ce, de)
            cds = max(cs, ds)
            if cde < ts:
                continue
            if cds > te:
                continue
            number_of_nt_deleted_in_exons += min(cde, te) - max(cds, ts) + 1

        if number_of_nt_deleted_in_exons == 0:
            stmt = 'select pos from splicetable where pos=' + str(de)
            cursor.execute(stmt)
            results = cursor.fetchall()
            if len(results) > 0:
                sequence_ontology = 'SS'
            else:
                sequence_ontology = ''
        elif number_of_nt_deleted_in_exons % 3 == 0:
            sequence_ontology = 'ID' # Inframe deletion
        else:
            sequence_ontology = 'FD' # Frameshift deletion

        return sequence_ontology

    def __set_sequence_ontology(self, codon, genomic):
        # Assumes new format.
        try:
            ref_seq_len = len(genomic.ref)
            alt_seq_len = len(genomic.alt)
            codon_uid = str(codon.UID)
            if genomic.ref == '-':
                if alt_seq_len % 3 == 0:
                    codon.sequence_ontology = 'II' # Inframe insertion
                else:
                    codon.sequence_ontology = 'FI' # Frameshift insertion
            elif genomic.alt == '-':
                codon.sequence_ontology = self.__get_deletion_sequence_ontology(codon_uid, genomic.chrom, genomic.strand, int(genomic.end), ref_seq_len)
            else:
                if ref_seq_len == 1 and alt_seq_len == 1:
                    ref_aa = codon.ref_aa
                    alt_aa = codon.alt_aa
                    if ref_aa == '*':
                        if alt_aa == '*':
                            codon.sequence_ontology = 'SY' #'synonymous_variant'
                        else:
                            codon.sequence_ontology = 'SL' #'stop_lost'
                    elif alt_aa == '*':
                        if ref_aa == '*':
                            codon.sequence_ontology = 'SY' #'synonymous_variant'
                        else:
                            codon.sequence_ontology = 'SG' #'stop_gained'
                    else:
                        if ref_aa == alt_aa:
                            codon.sequence_ontology = 'SY' #'synonymous_variant'
                        else:
                            codon.sequence_ontology = 'MS' #'missense_variant'
                else:
                    codon.sequence_ontology = 'CS'
        except Exception, e:
            raise

    def __pick_representative_codon(self, codons):
        try:
            if len(codons) == 0:
                return None

            best_transcript_codon = codons[0]
            for codon in codons:
                transcript_mrna_acc_type = codon.mrna_acc_type
                best_transcript_mrna_acc_type = best_transcript_codon.mrna_acc_type

                if codon.sequence_ontology == 'SS' and best_transcript_codon.sequence_ontology != 'SS':
                    continue
                if transcript_mrna_acc_type == 'Ensembl':
                    if best_transcript_mrna_acc_type != 'Ensembl':
                        best_transcript_codon = codon
                    else:
                        if codon.aa_len > best_transcript_codon.aa_len:
                            best_transcript_codon = codon
                elif transcript_mrna_acc_type == 'RefSeq':
                    if best_transcript_mrna_acc_type == 'Ensembl':
                        pass
                    elif best_transcript_mrna_acc_type == 'RefSeq':
                        if codon.aa_len > best_transcript_codon.aa_len:
                            best_transcript_codon = codon
                    elif best_transcript_mrna_acc_type == 'CCDS':
                        best_transcript_codon = codon
                elif transcript_mrna_acc_type == 'CCDS':
                    if best_transcript_mrna_acc_type in ['RefSeq', 'Ensembl']:
                        pass
                    elif best_transcript_mrna_acc_type == 'CCDS':
                        if codon.aa_len > best_transcript_codon.aa_len:
                            best_transcript_codon = codon
            return best_transcript_codon
        except Exception, e:
            raise

    def __get_codons(self, genomic):
        "Look up codons overlapping this genomic coordinate"

        try:
            genomic_uid_ref_base_checked = []
            codon = None
            codons = []
            uid_ref_match = True        #By default this is True
            discovered_transcripts = {} # To prevent duplicate transcripts from CodonTable and splicetable.
            for row in SpliceTblRow.query(self.conn, genomic):
                codon = Codon(row, splice=True)
                cursor = self.conn.cursor()
                cursor.execute('select RefseqT, EnsT, CCDS, aaLen ' +\
                               'from Transcript where UID=' + str(codon.UID))
                result = cursor.fetchone()
                if result != None:

                    aapos = row.get('aapos')
                    if aapos == -1:
                        continue

                    (refseqt, enst, ccds, aalen) = result
                    if refseqt != None:
                        mrnaAcc = refseqt
                        mrna_acc_type = 'RefSeq'
                    elif enst != None:
                        mrnaAcc = enst
                        mrna_acc_type = 'Ensembl'
                    elif ccds != None:
                        mrnaAcc = ccds
                        mrna_acc_type = 'CCDS'

                    if discovered_transcripts.has_key(mrnaAcc):
                        continue

                    codon.aa_len = aalen
                    codon.refseq = refseqt
                    codon.ens = enst
                    codon.ccds = ccds
                    codon.chrom = genomic.chrom
                    codon.mrnaAcc = mrnaAcc
                    codon.mrna_acc_type = mrna_acc_type
                    codon.codon_pos = -1
                    codon.codon_change = ''
                    codon.codonNo = aapos
                    codon.ref_codon = ''
                    codon.alt_codon = ''
                    codon.ref_aa = ''
                    codon.alt_aa = ''
                    codon.sequence_ontology = 'SS'
                    codon.pos1 = -1
                    codon.pos2 = -1
                    codon.pos3 = -1
                    codon.nucs = ''
                    codon.end = genomic.end
                    if discovered_transcripts.has_key(mrnaAcc) == False:
                        codons.append(codon)
                        discovered_transcripts[mrnaAcc] = True

            for row in CodonTblRow.query(self.conn, genomic):
                try:
                    codon = Codon(row)
                    if genomic.ref == '-' and (codon.strand != genomic.strand):
                        if genomic.strand == '-':
                            true_end = str(int(genomic.original_end) + 1)
                        else:
                            true_end = str(int(genomic.original_end) - 1)
                        genomic.end = true_end
                        if int(true_end) not in [codon.pos1, codon.pos2, codon.pos3]:
                            if codon.strand == '+':
                                if int(true_end) > codon.pos3:
                                    true_codonNo = codon.codonNo + 1
                                else:
                                    true_codonNo = codon.codonNo - 1
                            else:
                                if int(true_end) < codon.pos3:
                                    true_codonNo = codon.codonNo + 1
                                else:
                                    true_codonNo = codon.codonNo - 1
                            cursor = self.conn.cursor()
                            q = 'select t.aaLen as aa_len, t.RefseqT as refseq, t.EnsT as ens, t.CCDS as ccds, ct.* from '+\
                                '(select * from CodonTable where UID = %s and Pos = %s) '%(codon.UID, true_codonNo)+\
                                'as ct, Transcript as t where ct.UID=t.UID;'
                            cursor.execute(q)
                            qr = cursor.fetchone()
                            if qr:
                                codon = Codon(CodonTblRow(qr, cursor.description))
                            else:
                                continue
                    else:
                        #genomic.end = genomic.original_end
                        pass

                    if discovered_transcripts.has_key(codon.mrnaAcc):
                        continue

                    codon.end = genomic.end
                    self.__set_codon_pos(genomic, codon)
                    if (genomic_uid_ref_base_checked.count(genomic.uid) == 0):
                        uid_ref_match = self.__check_if_ref_input_correct(genomic, codon)
                        genomic_uid_ref_base_checked.append(genomic.uid)
                    if uid_ref_match == True:
                        self.__set_ref_alt_codons(codon, genomic)
                        self.__set_ref_alt_aas(codon, genomic)
                        self.__set_sequence_ontology(codon, genomic)
                        if codon.sequence_ontology != '':
                            codons.append(codon)
                            discovered_transcripts[codon.mrnaAcc] = True
                except Exception, e:
                    raise
                    sys.stderr.write(genomic.uid + '\t' + 'Codon conversion error\n')

            best_codon = None
#             best_codon by defalt will be None but IF the input ref matches the database ref then the best_codon will be picked from the codons detected
            if uid_ref_match == True:
                best_codon = self.__pick_representative_codon(codons)
            no_transcripts_for_genomic = len(codons)
            if self.pickOneTranscript == True:
                codons = [best_codon]

            return (codons, best_codon, no_transcripts_for_genomic, uid_ref_match)
        except Exception, e:
            raise

    def __isAltSplice(self, codons):
        "If at least 2 codons are split across an intron, case may be alt splice"
        try:
            altSplice = [x for x in codons.values() if x.size > 2]
            return len(altSplice) >= 1
        except Exception, e:
            raise

    def __selectCodon(self, codons, genomic):
        "Arbitrarily select codon - prioritize missense"
        try:
            codonlist = []
            for codon in codons:
                alt_base_pos_in_codon = self.__get_codonsPos(genomic, codon)
                if self.__matches(genomic, codon):
                    wtAA = self.__getRefAA(codon)
                    pos = codon.codonNo
                    altAA = self.__mutate(genomic, codon, alt_base_pos_in_codon)
                    if wtAA != None and altAA != None:
                        if self.__isMissense(wtAA, altAA):
                            codonlist.append(codon)
            if len(codonlist) < 1:
                codonlist = codons
            # Commented by hcarter on 06/28/2009 - although less biased, the random selection can
            # cause different mappings to be returned each time snvGet is run. Consistency is
            # more reproducible and easier to deal with computationally.
            #codon = random.sample(codonlist,1)[0]
            #return codon
            return codonlist[0]
        except Exception, e:
            raise

    def __getCodonPos(self, genomic, codon):
        """Determine position in the codon"""
        try:
            return [int(codon.pos1), int(codon.pos2), int(codon.pos3)].index(genomic.end)
        except Exception, e:
            raise

    def __set_codon_pos(self, genomic, codon):
        """Determine position in the codon"""
        try:
            codon.codon_pos = -1
            genomic_pos = int(genomic.end)
            if genomic_pos == codon.pos1:
#                 if genomic.ref == '-':
#                     codon.codon_pos = -1
#                 else:
#                     codon.codon_pos = 0
                codon.codon_pos = 0
            elif genomic_pos == codon.pos2:
                codon.codon_pos = 1
            elif genomic_pos == codon.pos3:
                codon.codon_pos = 2
        except Exception, e:
            raise

    def __matches(self, genomic, codon):
        """Make sure that supplied reference base matches documented reference base"""
        try:
            return (genomic.strand == codon.strand and genomic.ref == codon.nucs[codon.codon_pos]) or (genomic.strand != codon.strand and rcDict.get(genomic.ref) == codon.nucs[codon.codon_pos])
        except Exception, e:
            raise
    def __translate_codons(self, bases): #Currently, this silently drops extra bases beyond the last group of three, should this be handled differently?
        try:
            bases = bases.upper()
            aas = ''
            for i, nuc in enumerate(bases):
                if i%3 == 2:
                    codon_bases = ''.join(bases[i-2:i+1])
                    aas += codonTable.get(codon_bases, '?')
            return aas
        except Exception, e:
            raise

    def __translate_bases(self, bases):
        # This function will return the translated chain, and also the extra bases.
        # Compare to self.__translate_codons, which silently drops the extra bases.
        # Perhaps this should be default behavior?
        try:
            full_seq_length = len(bases)
            num_extra_bases = full_seq_length % 3
            codon_bases = bases[:-num_extra_bases]
            extra_bases = bases[-num_extra_bases:]
            aas = self.__translate_codons(codon_bases) # could integrate self.__translate_codons() code here if wanted to make this default translation method
            return aas, extra_bases
        except Exception, e:
            raise

    def __getRefAA(self, codon):
        """Return amino acid before base substitution """
        try:
            return codonTable.get(codon.nucs)
        except Exception, e:
            raise

    def __get_alt_codon(self, genomicVar, codon, alt_base_alt_base_pos_in_codon):
        try:
            alt_base = genomicVar.alt
            if genomicVar.strand != codon.strand:
                alt_base = rcDict.get(alt_base)
            alt_codon = codon.nucs[:alt_base_alt_base_pos_in_codon] + alt_base + codon.nucs[alt_base_alt_base_pos_in_codon+1:]
            return alt_codon
        except Exception, e:
            raise

    def __mutate(self, genomicVar, codon, pos):
        """Return amino acid after base substitution """
        try:
            alt = genomicVar.alt
            if genomicVar.strand != codon.strand:
                alt = rcDict.get(alt)
            altCodon = codon.nucs[:pos] + alt + codon.nucs[pos+1:]
            return codonTable.get(altCodon)
        except Exception, e:
            raise

    def __getMutationType(self):
        try:
            __getCdsMutationType(wtAA, altAA)
        except Exception, e:
            raise

    def __getCdsMutationType(self, wtAA, altAA):
        """Returns coding sequence mutation type using Sequence Ontology terms """
        try:
            if wtAA == '*':
                if altAA == '*':
                    mutationType = 'SY' #'synonymous_variant'
                else:
                    mutationType = 'SL' #'stop_lost'
            elif altAA == '*':
                if wtAA == '*':
                    mutationType = 'SY' #'synonymous_variant'
                else:
                    mutationType = 'SG' #'stop_gained'
            else:
                if wtAA == altAA:
                    mutationType = 'SY' #'synonymous_variant'
                else:
                    mutationType = 'MS' #'missense_variant'
            return mutationType
        except Exception, e:
            raise

class TranscriptMap(object):
    """
    Return a transcript and codon if it overlaps the genomic coordinate.
    """

    def __init__(self, dbconn):
        try:
            self.conn = dbconn
        except Exception, e:
            raise

    def lookup(self, var):
        try:
            return CodonTblRow.transcriptquery(self.conn, var)
        except Exception, e:
            raise

class DbRowBase(object):
    "base class used to convert a row of a query to an object"
    def __init__(self, row, description):
        try:
            for i in xrange(len(description)):
                if row != None:
                    self.__dict__[description[i][0]] = row[i]
                else:
                    self.__dict__[description[i][0]] = "NA"
        except Exception, e:
            raise

    def get(self, colname):
        try:
            return self.__dict__.get(colname)
        except Exception, e:
            raise

class SpliceTblRow(DbRowBase):
    @staticmethod

    def query(conn, var):
        try:
            cursor=conn.cursor()
            reflen = len(var.ref)
            altlen = len(var.alt)
            varend = var.end
            if reflen > 1 and altlen == 1: # Deletion starting in an intron
                if var.strand == '+':
                    de = varend + reflen - 1
                else:
                    de = varend - reflen + 1
                de = str(de)
                stmt = 'select UID, strand, aapos from splicetable ' + \
                       'where chr="' + var.chrom + '" and (pos=' + var.end + ' or ' +\
                       'pos=' + de + ')'
                cursor.execute(stmt)
            else:
                cursor.execute('select UID, strand, aapos from splicetable ' +
                               'where chr="' + var.chrom + '" and pos=' + var.end)
            while True:
                row = cursor.fetchone()
                (transcript_uid) = row
                if row == None:
                    break
                yield(SpliceTblRow(row, cursor.description))
        except Exception, e:
            raise

class CodonTblRow(DbRowBase):

    @staticmethod
    def query(conn, var):
        try:
            cursor=conn.cursor()
            cursor.execute('select t.aaLen as aa_len, ' +\
                           't.RefseqT as refseq, t.EnsT as ens, t.CCDS as ccds, ct.* ' +\
                           'from (' +\
                           'select * from CodonTable ' +\
                           'where chrom="' + var.chrom + '" and pos1=' + var.end + ' UNION ' +\
                           'select * from CodonTable ' +\
                           'where chrom="' + var.chrom + '" and pos2=' + var.end + ' UNION ' +\
                           'select * from CodonTable ' +\
                           'where chrom="' + var.chrom + '" and pos3=' + var.end +\
                           ') as ct, Transcript as t where ct.UID=t.UID')
            while True:
                row = cursor.fetchone()
                if row == None:
                    break
                yield(CodonTblRow(row, cursor.description))
        except Exception, e:
            raise

    @staticmethod
    def query_mutation(conn, var):
        try:
            cursor=conn.cursor()
            cursor.execute('SELECT t.aaLen as aa_len, ' +\
                           't.RefseqT as refseq, t.EnsT as ens, t.CCDS as ccds, ct.* ' +\
                           'FROM CodonTable as ct, Transcript as t '
                           "WHERE ct.UID=t.UID AND "
                           "(t.CCDS='{0}' OR t.RefseqT='{0}' OR t.EnsT='{0}') AND "
                           "ct.Pos={1}".format(var.acc, var.aa_pos))
            while True:
                row = cursor.fetchone()
                if row == None:
                    break
                yield(CodonTblRow(row, cursor.description))
        except Exception, e:
            raise

    @staticmethod
    def get_codon_bases(conn, uid, codon_no):
        try:
            cursor = conn.cursor()
            cursor.execute('select bases from CodonTable where UID=' + str(uid) + ' and Pos=' + str(codon_no))
            row = cursor.fetchone()
            if row == None:
                return None
            else:
                bases = row[0]
                return bases
        except Exception, e:
            raise

    @staticmethod
    def get_bases_in_chunks(conn, uid, codon_start, chunk_len):
        try:
            cursor = conn.cursor()
            complete_chunk = True
            while complete_chunk:
                codon_end = codon_start + chunk_len - 1
                cursor.execute('select bases from CodonTable where UID=%s and Pos between %s and %s;' \
                               %(uid, codon_start, codon_end))
                codon_start = codon_end + 1
                rows = cursor.fetchall()
                if len(rows) < chunk_len:
                    complete_chunk = False
                bases = ''.join([row[0] for row in rows])
                yield bases, complete_chunk
        except Exception, e:
            raise

    @staticmethod
    def transcriptquery(conn, var):
        try:
            cursor=conn.cursor()
            if var.chrom[0] != 'c':
                var.chrom = 'chr' + var.chrom.upper()
            sql = 'Select UID from CodonTable' +\
                ' where chrom="' + var.chrom + '" and pos1=' + var.end +\
                ' union ' +\
                'select UID from CodonTable' +\
                ' where chrom="' + var.chrom + '" and pos2=' + var.end +\
                ' union ' +\
                'select UID from CodonTable' +\
                ' where chrom="' + var.chrom + '" and pos3=' + var.end
            cursor.execute(sql)
            row = cursor.fetchone()
            if row == None:
                transcript = None
            else:
                uid = row[0]
                cursor.execute('select CCDS, RefseqT, EnsT from Transcript where UID=' + str(uid))
                (ccds, refseq_t, ens_t) = cursor.fetchone()
                if ens_t != None:
                    transcript = ens_t
                elif refseq_t != None:
                    transcript = refseq_t
                elif ccds != None:
                    transcript = ccds
            return transcript
        except Exception, e:
            raise


class Codon(object):
    """Note: pos1, pos2, pos3 are 0-based"""
    def __init__(self, row, splice=False):
        try:
            if splice:
                self.UID = row.get('UID')
                self.strand = row.get('strand')
                return
            self.UID = row.get('UID')
            self.aa_len = row.get('aa_len')
            self.refseq = row.get('refseq')
            self.ens = row.get('ens')
            self.ccds = row.get('ccds')
            if self.refseq != None:
                self.mrnaAcc = self.refseq
                self.mrna_acc_type = 'RefSeq'
            elif self.ens != None:
                self.mrnaAcc = self.ens
                self.mrna_acc_type = 'Ensembl'
            elif self.ccds != None:
                self.mrnaAcc = self.ccds
                self.mrna_acc_type = 'CCDS'
            self.codonNo = row.get("Pos")
            self.chrom = row.get("chrom")
            self.pos1 = int(row.get("pos1"))
            self.pos2 = int(row.get("pos2"))
            self.pos3 = int(row.get("pos3"))
            self.nucs = row.get("bases")
            if self.pos1 < self.pos3: # Forward strand
                self.strand = "+"
                self.size = self.pos3-self.pos1
            else: # Reverse Strand
                self.strand = "-"
                self.size = self.pos1-self.pos3
            self.aa = codonTable[self.nucs]
        except Exception, e:
            raise

    def __str__(self):
        try:
            return str(self.UID) + ',' +\
                   self.mrnaAcc + ',' +\
                   str(self.aa_len) + ',' +\
                   str(self.codonNo) + ',' +\
                   self.ref_aa + ',' +\
                   self.alt_aa + ',' +\
                   self.chrom + ',' +\
                   str(self.end) + ',' +\
                   self.strand + ',' +\
                   str(self.pos1) + ',' +\
                   str(self.pos2) + ',' +\
                   str(self.pos3) + ',' +\
                   self.nucs + ',' +\
                   str(self.codon_pos)
        except Exception, e:
            raise
