from util import GenomicCoord

class TSVwriter(object):
    """Object to write tab-delimited files in format useable by most machine
    learning packages."""

    def __init__(self, name):
        self.handle = open(name,"w")
        self.features = ['UID', "ID"]

    def addFeature(self, feature):
        """Add feature to list for header column."""
        self.features.append(feature)

    def addClass(self, classlabels, weights=None):
        if classlabels is not None:
            self.features.append('class')

    def dataStart(self):
        """Write the header."""
        self.handle.write('\t'.join(self.features)+'\n')

    def addData(self, uid, name, values, classlabel=None):
        """Add feature row for mutation."""
        try:
            if values != None:
                if classlabel != None:
                    outstr = str(uid) + "\t" + name + "\t" + "\t".join(map(str,values)) + "\t" + classlabel + "\n"
                    self.handle.write(outstr)
                else:
                    outstr = str(uid) + "\t" + name + "\t" + "\t".join(map(str,values)) + "\n"
                    self.handle.write(outstr)
        except Exception as e:
            raise

    def close(self):
        """Close file handle."""
        self.handle.close()


class TsvReader(object):

    def __init__(self, filename):
        self.fh = file(filename, 'r')

    def __iter__(self):
        return self

    def next(self):
        while True:
            line = self.fh.readline()
            if line == "":
                self.fh.close()
                raise StopIteration
            line = line[:-1]
            if line.find("\t") == -1:
                raise Exception("Expected a tab delimited file.")
            return GenomicCoord(line.split("\t"))
