#External lib imports
import sys
import random
import string
import time

class DbRowBase(object):
    "base class used to convert a row of a query to an object"
    def __init__(self):
        pass

    def __init__(self, row, description):
        for i in xrange(len(description)):
            if row != None:
                self.__dict__[description[i][0]] = row[i] if row[i] != None else 0.0
            else:
                self.__dict__[description[i][0]] = None

    def get(self, colname):
        return self.__dict__.get(colname)


class TranscriptFeatures(DbRowBase):
    ""

    @staticmethod
    def query(conn, acc):
        try:
            query = TranscriptFeatures.buildQuery(acc)
            cursor = conn.cursor()
            cursor.execute(query)
            row = cursor.fetchone()
            feats = TranscriptFeatures(row,cursor.description)
            if row != None:
                row = cursor.fetchone()
                if row != None:
                    pass
                    # Commented by HC (11/18/2012)
                    # Can have multiple transcripts with the same CDS so multiple
                    # query matches should be allowed.
                    # raise Exception("Unexpected duplicate hit in database\n")
                cursor.close()
            return feats
        except Exception, e:
            raise

    @staticmethod
    def buildQuery(acc):
        try:
            '''
            if acc.find(".") != -1:
                core = acc.split(".")[0]
                ver = acc.split(".")[1]
            else:
                core = acc
                ver = None
            '''
            if acc.startswith("NP"):
                return "Select UID, RefseqP as tsID, aaLen from Transcript where RefseqP=\"" + acc + "\";"
            elif acc.startswith("NM"):
                return "Select UID, RefseqT as tsID, aaLen from Transcript where RefseqT=\"" + acc + "\";"
            elif acc.startswith("ENSP"):
                return "Select UID, EnsP as tsID, aaLen from Transcript where EnsP=\""+ acc +"\";"
            elif acc.startswith("ENST"):
                return "Select UID, EnsT as tsID, aaLen from Transcript where EnsT=\""+ acc +"\";"
            elif acc.startswith("CCDS"):
                return "Select UID, CCDS as tsID, aaLen from Transcript where CCDS=\""+ acc +"\";"
            else:
                raise Exception(acc + " not recognized. Accession may not supported by SNVbox\n")
        except Exception, e:
            raise

class AAFeatures(DbRowBase):
    ""

    @staticmethod
    def query(conn, features, wt, alt):
        try:
            cursor = conn.cursor()
            if wt == '' or alt == '':
                cursor.execute("Select " + ",".join(features) + " from AA_Features limit 1;")
                feats = AAFeatures(None, cursor.description)
            else:
                cursor.execute("Select " + ",".join(features) + " from AA_Features where WildType=\"" + wt+ "\" and Mut = \"" + alt + "\";")
                row = cursor.fetchone()
                feats = AAFeatures(row,cursor.description)
                if row != None:
                    row = cursor.fetchone()
                    if row != None:
                        sys.stderr.write("AAFeatures:Duplicate hit:"+stmt+"\n")
            cursor.close()
            return feats
        except Exception, e:
            raise

class RCFeatures(DbRowBase):
    ""

    @staticmethod
    def query(conn, features, UID, pos):
        try:
            cursor = conn.cursor()
            cursor.execute("Select " + ",".join(features) + " " +\
                           "from Regional_Comp " +\
                           "where UID = \"" + UID + "\" and Pos= " + pos+ ";")
            row = cursor.fetchone()
            feats = RCFeatures(row,cursor.description)
            if row != None:
                row = cursor.fetchone()
                if row != None:
                    sys.stderr.write("RCFeatures:Duplicate hit:"+stmt+"\n")
            cursor.close()
            return feats
        except Exception, e:
            raise

class LSFeatures(DbRowBase):
    ""

    @staticmethod
    def query(conn, features, UID, pos):
        try:
            cursor = conn.cursor()
            stmt = 'select ' + ','.join(features) + \
                           ' from Local_Structure local ' + \
                           ' where local.UID=' + UID + ' and local.Pos=' + pos
            cursor.execute(stmt)
            result = cursor.fetchone()
            if result != None:
                feats = LSFeatures(result,cursor.description)
                row = cursor.fetchone()
                if row != None:
                    sys.stderr.write("LSFeatures:Duplicate hit:"+stmt+"\n")
            else:
                feats = LSFeatures(None, cursor.description)
            cursor.close()
            return feats
        except Exception, e:
            raise

class GMSAFeatures(DbRowBase):
    ""

    @staticmethod
    def query(conn, features, UID, pos,  alt):
        try:
            cursor = conn.cursor()
            q_features = [x for x in features if x != "PHC"]
            if alt == '':
                if "PHC" in features:
                    q_features.append("PHC_A as PHC")
            else:
                if "PHC" in features:
                    q_features.append("PHC_" + alt + " as PHC")
            cursor.execute("Select " + ",".join(q_features) + " " +\
                           "from Genomic_MSA " +\
                           "where UID = \"" + UID + "\" and Pos= " + pos+ ";")
            row = cursor.fetchone()
            feats = GMSAFeatures(row,cursor.description)
            if row != None:
                row = cursor.fetchone()
                if row != None:
                    #sys.stderr.write("GMSAFeatures:Duplicate hit:"+stmt+"\n")
                    sys.stderr.write("GMSAFeatures:Duplicate hit:\n")
            cursor.close()
            return feats
        except Exception, e:
            raise

class SMSAFeatures(DbRowBase):
    ""

    @staticmethod
    def query(conn, features, UID, pos, alt):
        try:
            cursor = conn.cursor()
            q_features = [x for x in features if x != "PHC"]
            if alt == '':
                if "PHC" in features:
                    q_features.append("PHC_A as PHC")
            else:
                if "PHC" in features:
                    q_features.append("PHC_" + alt + " as PHC")
            cursor.execute('select ' + ','.join(q_features) + ' from Sam_MSA sam where sam.UID=' + UID + ' and sam.Pos=' + pos)
            result = cursor.fetchone()
            feats = SMSAFeatures(result,cursor.description)
            if result != None:
                row = cursor.fetchone()
                if row != None:
                    sys.stderr.write("SMSAFeatures:Duplicate hit:"+stmt+"\n")
            cursor.close()
            return feats
        except Exception, e:
            raise

class ExonFeatures(DbRowBase):
    ""

    @staticmethod
    def query(conn, features, UID, pos):
        try:
            cursor = conn.cursor()
            q_features = ["e." + x for x in features]
            stmt = "Select " + ",".join(q_features) + " " +\
                   'from Transcript_Exon t, Exon_Features e ' +\
                   "where t.aaStart <= " + str(pos) + \
                   " and t.aaEnd >= " + str(pos) + " " +\
                   "and t.UID=e.UID and t.Exon=e.Exon and t.UID=" + UID + ";"
            cursor.execute(stmt)
            row = cursor.fetchone()
            feats = ExonFeatures(row,cursor.description)
            if row != None:
                row = cursor.fetchone()
                if row != None:
                    sys.stderr.write("ExonFeatures:Duplicate hit:"+stmt+"\n")
            cursor.close()
            return feats
        except Exception, e:
            raise

class UniprotFeatures(DbRowBase):
    ""

    @staticmethod
    def query(conn, features, UID, pos):
        try:
            cursor = conn.cursor()
            q_features = ["u." + x for x in features]
            stmt = "Select " + ",".join(q_features) +\
                           " from Uniprot_Xref x left " +\
                           "join Uniprot_features u " +\
                           "on u.Acc=x.Uniprot and u.Pos=x.UniprotPos " +\
                           "where x.UID=" + UID + " and x.Pos=" + pos + ";"
            cursor.execute(stmt)
            row = cursor.fetchone()
            feats = UniprotFeatures(row,cursor.description)
            if row != None:
                row = cursor.fetchone()
                if row != None:
                    sys.stderr.write("UniprotFeatures:Duplicate hit:"+stmt+"\n")
            cursor.close()
            return feats
        except Exception, e:
            raise

class TripletFeatures(DbRowBase):
    ""

    tripletFeaturesData = None

    @staticmethod
    def query(conn, triplet):
        try:
            cursor = conn.cursor()

            # Loads Triplet_Features table into memory.
            if TripletFeatures.tripletFeaturesData == None:
                TripletFeatures.tripletFeaturesData = {}
                cursor.execute('select * from Triplet_Features')
                results = cursor.fetchall()
                for result in results:
                    (ttt, firstprob, secondprob, thirdprob) = result
                    TripletFeatures.tripletFeaturesData[ttt] = [firstprob, secondprob, thirdprob]

            cursor.close()

            if TripletFeatures.tripletFeaturesData.has_key(triplet):
                return TripletFeatures.tripletFeaturesData[triplet]
            else:
                return None

        except Exception, e:
            raise

    @staticmethod
    def formatQuery(feature, UID, pos, res):
        try:
            if feature == "FirstProb":
                mkTriplet = "CONCAT(\""+res+"\", r1.WildType, r2.WildType) and r1.UID="+UID+" and r2.UID="+UID+" and r1.Pos="+pos+"+1 and r2.Pos="+pos+"+2"
            elif feature == "SecondProb":
                mkTriplet = "CONCAT(r1.WildType, \""+res+"\", r2.WildType) and r1.UID="+UID+" and r2.UID="+UID+" and r1.Pos="+pos+"+-1 and r2.Pos="+pos+"+1"
            elif feature == "ThirdProb":
                mkTriplet = "CONCAT(r1.WildType, r2.WildType, \""+res+"\") and r1.UID="+UID+" and r2.UID="+UID+" and r1.Pos="+pos+"-2 and r2.Pos="+pos+"-1"
            return mkTriplet
        except Exception, e:
            raise

class TripletSubFeatures(DbRowBase):
    ""

    tripletSubFeatures = None

    @staticmethod
    def query(conn, two, wt, mut):
        try:
            cursor = conn.cursor()

            # Loads the whole TripletSubFeatures table.
            if TripletSubFeatures.tripletSubFeatures == None:
                TripletSubFeatures.tripletSubFeatures = {}
                cursor.execute('select * from Triplet_Sub_Features')
                results = cursor.fetchall()
                for result in results:
                    (twobases, wildtype, mutant, firstdiffprob, seconddiffprob, thirddiffprob) = result
                    if TripletSubFeatures.tripletSubFeatures.has_key(twobases) == False:
                        TripletSubFeatures.tripletSubFeatures[twobases] = {}
                    if TripletSubFeatures.tripletSubFeatures[twobases].has_key(wildtype) == False:
                        TripletSubFeatures.tripletSubFeatures[twobases][wildtype] = {}
                    TripletSubFeatures.tripletSubFeatures[twobases][wildtype][mutant] = [firstdiffprob, seconddiffprob, thirddiffprob]

            cursor.close()

            if TripletSubFeatures.tripletSubFeatures.has_key(two):
                if TripletSubFeatures.tripletSubFeatures[two].has_key(wt):
                    if TripletSubFeatures.tripletSubFeatures[two][wt].has_key(mut):
                        return TripletSubFeatures.tripletSubFeatures[two][wt][mut]

            return [None, None, None]

        except Exception, e:
            raise

    @staticmethod
    def formatQuery(feature, UID, pos):
        try:
            if feature == "FirstDiffProb":
                mkTwoBases = "CONCAT(r1.WildType, r2.WildType) and r1.UID="+UID+" and r2.UID="+UID+" and r1.Pos="+pos+"+1 and r2.Pos="+pos+"+2"
            elif feature == "SecondDiffProb":
                mkTwoBases = "CONCAT(r1.WildType, r2.WildType) and r1.UID="+UID+" and r2.UID="+UID+" and r1.Pos="+pos+"+-1 and r2.Pos="+pos+"+1"
            elif feature == "ThirdDiffProb":
                mkTwoBases = "CONCAT(r1.WildType, r2.WildType) and r1.UID="+UID+" and r2.UID="+UID+" and r1.Pos="+pos+"-2 and r2.Pos="+pos+"-1"
            return mkTwoBases
        except Exception, e:
            raise

class AlaSubPhcFeatures (DbRowBase):

    @staticmethod
    def query(conn, UID, pos):
        try:
            cursor = conn.cursor()
            cursor.execute("Select AlaSubPhc " +\
                           "from AlaSubPhc " +\
                           "where UID = \"" + UID + "\" and Pos= " + pos+ ";")
            row = cursor.fetchone()
            feats = AlaSubPhcFeatures(row,cursor.description)
            if row != None:
                row = cursor.fetchone()
                if row != None:
                    sys.stderr.write("AlaSubPhcFeatures:Duplicate hit:"+stmt+"\n")
            cursor.close()
            return feats
        except Exception, e:
            raise

class IndelFeatures (DbRowBase):

    @staticmethod
    def query(conn, var):
        aa_len = int(var.aa_len)
        aa_pos = int(var.aa_pos)
        remaining = float(aa_len - aa_pos)/float(aa_len)
        reflen = None
        altlen = None
        if hasattr(var, 'refbase'):
            if var.refbase == '-':
                reflen = 0
            else:
                reflen = len(var.refbase)
        else:
            reflen = "?"
        if hasattr(var, 'altbase'):
            if var.altbase == '-':
                altlen = 0
            else:
                altlen = len(var.altbase)
        else:
            altlen = "?"

        if hasattr(var, 'transcript_count'):
            transcript_count = var.transcript_count
        else:
            transcript_count = "?"
        cursor = conn.cursor()

        hugo = ''
        if hasattr(var, 'hugo') == False:
            cursor.execute('select GeneSymbol from GeneSymbols where UID=' + str(var.UID))
            result = cursor.fetchone()
            if result != None:
                hugo = result[0]
        else:
            hugo = var.hugo

        cmd = 'select logn from Indel_Features where hugo="' + hugo + '"'
        cursor.execute(cmd)



