#External lib imports
import sys
import random
import string
import time
from db_row import *


class DbTableBase(object):
    """Base class for storing feature information retrieved from SNVBOX tables"""
    def __init__(self):
        try:
            self.featurelist = []
            self.featureValues = []
            self.featurenames = []
            self.featureMean = {}
            self.featureRMS = {}
        except Exception, e:
            raise

class AATable(DbTableBase):

    def getFeatures(self, conn, var):
        try:
            self.featureValues = []
            row = AAFeatures.query(conn, self.featurelist, var.ref_aa, var.alt_aa)
            for feature in self.featurelist:
                self.featureValues.append(row.get(feature))
            return self.featureValues
        except Exception, e:
            raise

class RCTable(DbTableBase):

    def getFeatures(self, conn, var):
        try:
            self.featureValues = []
            row = RCFeatures.query(conn, self.featurelist, var.UID, var.aa_pos)
            for feature in self.featurelist:
                self.featureValues.append(row.get(feature))
            return self.featureValues
        except Exception, e:
            raise

class LSTable(DbTableBase):

    def getFeatures(self, conn, var):
        try:
            self.featureValues = []
            row = LSFeatures.query(conn, self.featurelist, var.UID, var.aa_pos)
            for feature in self.featurelist:
                self.featureValues.append(row.get(feature))
            return self.featureValues
        except Exception, e:
            raise

class GMSATable(DbTableBase):

    def getFeatures(self, conn, var):
        try:
            self.featureValues = []
            row = GMSAFeatures.query(conn, self.featurelist, var.UID, var.aa_pos, var.alt_aa)
            for feature in self.featurelist:
                self.featureValues.append(row.get(feature))
            return self.featureValues
        except Exception, e:
            raise

class SMSATable(DbTableBase):

    def getFeatures(self, conn, var):
        try:
            self.featureValues = []
            row = SMSAFeatures.query(conn, self.featurelist, var.UID, var.aa_pos, var.alt_aa)
            for feature in self.featurelist:
                self.featureValues.append(row.get(feature))
            return self.featureValues
        except Exception, e:
            raise

class ExonTable(DbTableBase):

    def getFeatures(self, conn, var):
        try:
            self.featureValues = []
            row = ExonFeatures.query(conn, self.featurelist, var.UID, var.aa_pos)
            for feature in self.featurelist:
                self.featureValues.append(row.get(feature))
            return self.featureValues
        except Exception, e:
            raise

class UniprotTable(DbTableBase):

    def getFeatures(self, conn, var):
        try:
            self.featureValues = []
            row = UniprotFeatures.query(conn, self.featurelist, var.UID, var.aa_pos)
            for feature in self.featurelist:
                self.featureValues.append(row.get(feature))
            return self.featureValues
        except Exception, e:
            raise

class AlaSubPhcTable(DbTableBase):

    def getFeatures(self, conn, var):
        try:
            self.featureValues = []
            row = AlaSubPhcFeatures.query(conn, var.UID, var.aa_pos)
            for feature in self.featurelist:
                self.featureValues.append(row.get(feature))
            return self.featureValues
        except Exception, e:
            raise

class IndelTable(DbTableBase):

    def getFeatures(self, conn, var):
        try:
            self.featureValues = []
            row = IndelFeatures.query(conn, var)
            for feature in self.featurelist:
                self.featureValues.append(row.get(feature))
            return self.featureValues
        except Exception, e:
            raise


class Triplets:
    uid = None
    pos = None
    triplets = None

    @staticmethod
    def findTriplets(conn, var):
        try:
            cursor = conn.cursor()

            Triplets.uid = var.UID
            Triplets.pos = var.aa_pos

            # Gets the three triplets (having the mutation amino acid in the first, second, and third position).
            first_triplet = None
            second_triplet = None
            third_triplet = None
            pos = int(var.aa_pos)
            pos_m2 = pos - 2
            pos_m1 = pos - 1
            pos_p1 = pos + 1
            pos_p2 = pos + 2

            cursor.execute('select Pos, WildType from Regional_Comp where UID=' + var.UID + ' and Pos>=' + str(pos_m2) + ' and Pos<=' + str(pos_p2))
            results = cursor.fetchall()
            pos_to_aa = {}
            for result in results:
                (pos, aa) = result
                pos_to_aa[pos] = aa

            if pos_to_aa.has_key(pos_m2) == False or pos_to_aa.has_key(pos_m1) == False or pos_to_aa.has_key(pos) == False:
                triplet_3_wt = None
                triplet_3_mut = None
            else:
                triplet_3_wt = pos_to_aa[pos_m2] + pos_to_aa[pos_m1] + var.ref_aa
                triplet_3_mut = pos_to_aa[pos_m2] + pos_to_aa[pos_m1] + var.alt_aa

            if pos_to_aa.has_key(pos_m1) == False or pos_to_aa.has_key(pos) == False or pos_to_aa.has_key(pos_p1) == False:
                triplet_2_wt = None
                triplet_2_mut = None
            else:
                triplet_2_wt = pos_to_aa[pos_m1] + var.ref_aa + pos_to_aa[pos_p1]
                triplet_2_mut = pos_to_aa[pos_m1] + var.alt_aa + pos_to_aa[pos_p1]

            if pos_to_aa.has_key(pos) == False or pos_to_aa.has_key(pos_p1) == False or pos_to_aa.has_key(pos_p2) == False:
                triplet_1_wt = None
                triplet_1_mut = None
            else:
                triplet_1_wt = var.ref_aa + pos_to_aa[pos_p1] + pos_to_aa[pos_p2]
                triplet_1_mut = var.alt_aa + pos_to_aa[pos_p1] + pos_to_aa[pos_p2]

            Triplets.triplets = [triplet_1_wt, triplet_1_mut, triplet_2_wt, triplet_2_mut, triplet_3_wt, triplet_3_mut]
        except Exception, e:
            raise

class TripletTable(DbTableBase):

    def getFeatures(self, conn, var):
        try:
            if var.ref_aa == '' or var.alt_aa == '':
                self.featureValues = []
                for i in xrange(len(self.featurelist)):
                    self.featureValues.append(None)
            else:
                if Triplets.uid != var.UID or Triplets.pos != var.aa_pos:
                    Triplets.findTriplets(conn, var)
                [triplet_1_wt, triplet_1_mut, triplet_2_wt, triplet_2_mut, triplet_3_wt, triplet_3_mut] = Triplets.triplets

                self.featureValues = {}
                if triplet_1_mut == None:
                    prob = None
                else:
                    prob = TripletFeatures.query(conn, triplet_1_mut)
                    if prob != None:
                        prob = prob[0]
                self.featureValues['AATripletFirstProbMut'] = prob
                if triplet_1_wt == None:
                    prob = None
                else:
                    prob = TripletFeatures.query(conn, triplet_1_wt)
                    if prob != None:
                        prob = prob[0]
                self.featureValues['AATripletFirstProbWild'] = prob
                if triplet_2_mut == None:
                    prob = None
                else:
                    prob = TripletFeatures.query(conn, triplet_2_mut)
                    if prob != None:
                        prob = prob[1]
                self.featureValues['AATripletSecondProbMut'] = prob
                if triplet_2_wt == None:
                    prob = None
                else:
                    prob = TripletFeatures.query(conn, triplet_2_wt)
                    if prob != None:
                        prob = prob[1]
                self.featureValues['AATripletSecondProbWild'] = prob
                if triplet_3_mut == None:
                    prob = None
                else:
                    prob = TripletFeatures.query(conn, triplet_3_mut)
                    if prob != None:
                        prob = prob[2]
                self.featureValues['AATripletThirdProbMut'] = prob
                if triplet_3_wt == None:
                    prob = None
                else:
                    prob = TripletFeatures.query(conn, triplet_3_wt)
                    if prob != None:
                        prob = prob[2]
                self.featureValues['AATripletThirdProbWild'] = prob
            ret_values = []
            for feature in self.featurenames:
                ret_values.append(self.featureValues[feature])
            return ret_values
        except Exception, e:
            raise

class TripletSubTable(DbTableBase):

    def getFeatures(self, conn, var):
        try:
            if var.ref_aa == '' or var.alt_aa == '':
                self.featureValues = [None, None, None]
            else:
                if Triplets.uid != var.UID or Triplets.pos != var.aa_pos:
                    Triplets.findTriplets(conn, var)
                [triplet_1_wt, triplet_1_mut, triplet_2_wt, triplet_2_mut, triplet_3_wt, triplet_3_mut] = Triplets.triplets

                self.featureValues = {}
                if triplet_1_wt == None:
                    firstdiffprob = None
                else:
                    [firstdiffprob, seconddiffprob, thirddiffprob] = TripletSubFeatures.query(conn, triplet_1_wt[1:3], var.ref_aa, var.alt_aa)
                self.featureValues['AATripletFirstDiffProb'] = firstdiffprob
                if triplet_2_wt == None:
                    seconddiffprob = None
                else:
                    [firstdiffprob, seconddiffprob, thirddiffprob] = TripletSubFeatures.query(conn, triplet_2_wt[0] + triplet_2_wt[2], var.ref_aa, var.alt_aa)
                self.featureValues['AATripletSecondDiffProb'] = seconddiffprob
                if triplet_3_wt == None:
                    thirddiffprob = None
                else:
                    [firstdiffprob, seconddiffprob, thirddiffprob] = TripletSubFeatures.query(conn, triplet_3_wt[:2], var.ref_aa, var.alt_aa)
                self.featureValues['AATripletThirdDiffProb'] = thirddiffprob
            ret_values = []
            for feature in self.featurenames:
                ret_values.append(self.featureValues[feature])
            return ret_values
        except Exception, e:
            raise
