class MAFReader(object):
    """
    Read a file with genomic coordinates in MAF format
    """
    def __init__(self, filename):
        self.fh = file(filename, 'r')
        self.is_header = True
        self.header2ix = {}

    def __iter__(self):
        return self

    def next(self):
        while True:
            try:
                line = self.fh.readline()
                if line.startswith('#'):
                    # skip comment line
                    continue
                if line == "":
                    self.fh.close()
                    raise StopIteration
                line=line.strip()

                # parse header
                if self.is_header:
                    line_split = line.split('\t')
                    self.header2ix = {l: i for i, l in enumerate(line_split)}
                    self.is_header = False
                    continue

                return MAF(line, self.header2ix)
            except Exception as e:
                raise


class MAF(object):
    """
    Store genomic coordinate of MAF line
    """
    index =  1

    def __init__(self, line, header2ix):
        try:
            self.index = MAF.index
            MAF.index += 1
            self.line = line
            row = self.line.split("\t")
        except Exception as e:
            raise
        if len(row) > 0:
            try:
                self.uid = str(self.index)
                self.chrom = row[header2ix['Chromosome']]
                if not self.chrom.startswith('chr'):
                    self.chrom = 'chr' + self.chrom
                self.end = row[header2ix['End_Position']]
                self.strand = '+'
                self.ref = row[header2ix['Reference_Allele']]
                self.alt = row[header2ix['Tumor_Seq_Allele2']]
            except Exception as e:
                raise
        else:
            raise Exception("Expected a tab delimited file")
