class MutationReader(object):
    """
    Read file with transcript and
    amino acid substitutions
    """
    def __init__(self, filename, indel):
        self.fh = file(filename, 'r')
        self.indel = indel

    def __iter__(self):
        return self

    def next(self):
        while True:
            try:
                line = self.fh.readline()[:-1]
                if line == '':
                    self.fh.close()
                    raise StopIteration
                line = line.replace(' ', '\t')
                var = Mutation(line, self.indel)
                return var
            except Exception as e:
                raise

class Mutation(object):
    """
    Stores missense mutation
    """
    index = 1

    def __init__(self, line, indel=False):
        # Format of line:
        # uid <tab> mRNA accession <tab> codon change <tab> AA position <tab> reference AA <tab> alternate AA <tab> AA len
        #
        try:
            self.index = Mutation.index
            Mutation.index += 1
            self.line = line
            self.genomicstatus = 0
            self.ref_flag = ''
            if line.find("\t") != -1:
                toks = self.line.split('\t')
                if indel: # For indel files written by info.py
                    self.uid = toks[0]
                    self.acc = toks[1]
                    self.ref_aa = ''
                    self.aa_pos = toks[2]
                    self.alt_aa = ''
                    self.hugo = toks[3]
                    self.refbase = toks[4]
                    self.altbase = toks[5]
                    self.transcript_count = toks[6]
                    self.aa_len = toks[7]
                    self.UID = toks[8]
                    reference_flag = toks[9]
                    if reference_flag == '1':
                        self.reference_flag = True
                    else:
                        self.reference_flag = False
                else: # For other mutation type files
                    # If no UID is supplied and if transcript coordinates are used
                    if len(toks) == 2:
                        self.uid = str(self.index)
                        self.acc = toks[0]
                        self.ref_aa = toks[1][0]
                        self.aa_pos = toks[1][1:-1]
                        self.alt_aa = toks[1][-1]
                        self.UID = None
                        self.reference_flag = True
                    # if no codon change is supplied
                    elif len(toks) == 3:
                        self.uid = toks[0]
                        self.acc = toks[1]
                        self.ref_aa = toks[2][0]
                        self.aa_pos = toks[2][1:-1]
                        self.alt_aa = toks[2][-1]
                        self.UID = None
                        self.reference_flag = True
                    elif len(toks) == 4:
                        self.uid = toks[0]
                        self.acc = toks[1]
                        self.ref_aa = toks[2][0]
                        self.aa_pos = toks[2][1:-1]
                        self.alt_aa = toks[2][-1]
                        self.UID = None
                        reference_flag = toks[3]
                        if reference_flag == '1':
                            self.reference_flag = True
                        else:
                            self.reference_flag = False
                    elif len(toks) == 8:
                        self.uid = toks[0]
                        self.acc = toks[1]
                        self.codon_change = toks[2]
                        self.aa_pos = toks[3]
                        self.ref_aa = toks[4]
                        self.alt_aa = toks[5]
                        self.aa_len = toks[6]
                        self.codon_strand = toks[7]
                        self.UID = None
                        self.reference_flag = True
                    elif len(toks) == 9:
                        self.uid = toks[0]
                        self.acc = toks[1]
                        self.codon_change = toks[2]
                        self.aa_pos = toks[3]
                        self.ref_aa = toks[4]
                        self.alt_aa = toks[5]
                        self.aa_len = toks[6]
                        self.codon_strand = toks[7]
                        self.UID = None
                        reference_flag = toks[8]
                        if reference_flag == '1':
                            self.reference_flag = True
                        else:
                            self.reference_flag = False
            else:
                raise Exception("Expected a tab delimited file")
        except Exception as e:
            raise

    def storeGenomicCoords(self, genomic):
        """
        If generated from a genomic event by snvGetTranscriptList,
        store genomic coordinates
        """
        try:
            self.chrom = genomic.chrom
            self.end = genomic.end
            self.strand = genomic.strand
            self.refbase = genomic.ref
            self.altbase = genomic.alt
            self.genomicstatus = 1
        except Exception as e:
            raise

    def __repr__(self):
        try:
            if hasattr(self, 'codon_change'):
                return "\t".join([str(self.uid), self.acc, self.codon_change, self.aa_pos, self.ref_aa, self.alt_aa, str(self.UID), self.chrom, self.end, self.strand, self.refbase, self.altbase])
            else:
                return "\t".join([str(self.uid), self.acc, self.aa_pos, self.ref_aa, self.alt_aa, str(self.UID)])
        except Exception as e:
            raise

